/* dispatch.c

   Network input dispatcher... */

/*
 * Copyright (c) 2004-2006 by Internet Systems Consortium, Inc. ("ISC")
 * Copyright (c) 1995-2003 by Internet Software Consortium
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *   Internet Systems Consortium, Inc.
 *   950 Charter Street
 *   Redwood City, CA 94063
 *   <info@isc.org>
 *   http://www.isc.org/
 *
 * This software has been written for Internet Systems Consortium
 * by Ted Lemon in cooperation with Vixie Enterprises and Nominum, Inc.
 * To learn more about Internet Systems Consortium, see
 * ``http://www.isc.org/''.  To learn more about Vixie Enterprises,
 * see ``http://www.vix.com''.   To learn more about Nominum, Inc., see
 * ``http://www.nominum.com''.
 */

#ifndef lint
static char copyright[] =
"$Id: dispatch.c,v 1.63.2.8 2006/02/22 22:43:27 dhankins Exp $ Copyright (c) 2004-2006 Internet Systems Consortium.  All rights reserved.\n";
#endif /* not lint */

#include "dhcpd.h"

struct timeout {
#ifndef LIBDHCP
	struct timeout *next;
#endif
	TIME when;
	void (*func) PROTO ((void *));
	void *what;
	tvref_t ref;
	tvunref_t unref;
};

#ifdef LIBDHCP
static struct timeout *timeouts = NULL;
static int ntimeouts = 0;
#else
static struct timeout *timeouts = NULL;
static struct timeout *free_timeouts = NULL;
#endif

void set_time(TIME t)
{
	/* Do any outstanding timeouts. */
	if (cur_time != t) {
		cur_time = t;
		process_outstanding_timeouts ((struct timeval *)0);
	}
}

struct timeval *process_outstanding_timeouts (struct timeval *tvp)
{
#ifdef LIBDHCP
	int i;
	struct timeout t = { 0 };
#endif
	/* Call any expired timeouts, and then if there's
	   still a timeout registered, time out the select
	   call then. */
#ifdef LIBDHCP
	if (!ntimeouts)
		return NULL;

	for (i = 0; i < ntimeouts && timeouts[i].when <= cur_time;) {
		struct timeout *new_timeouts;
		size_t n;

		memmove(&t, &timeouts[i], sizeof (t));

		n = (ntimeouts - i - 1) * sizeof (t);
		memmove(&timeouts[i+1], &timeouts[i], n);

		n = --ntimeouts * sizeof (t);
		new_timeouts = realloc(timeouts, n);
		/* XXX broken API, no way to return error here */
		if (new_timeouts || !n)
			timeouts = new_timeouts;

		if (t.func)
			t.func(t.what);
		if (t.unref)
			t.unref(t.what, MDL);
	}
	if (tvp && ntimeouts) {
		tvp->tv_sec = timeouts[0].when;
		tvp->tv_usec = 0;
#else
      another:
	if (timeouts) {
		struct timeout *t;
		if (timeouts -> when <= cur_time) {
			t = timeouts;
			timeouts = timeouts -> next;
			(*(t -> func)) (t -> what);
			if (t -> unref)
				(*t -> unref) (&t -> what, MDL);
			t -> next = free_timeouts;
			free_timeouts = t;
			goto another;
		}
		if (tvp) {
			tvp -> tv_sec = timeouts -> when;
			tvp -> tv_usec = 0;
		}
#endif
		return tvp;
#ifdef LIBDHCP
	}
	return NULL;
#else
	} else
		return (struct timeval *)0;
#endif
}

/* Wait for packets to come in using select().   When one does, call
   receive_packet to receive the packet and possibly strip hardware
   addressing information from it, and then call through the
   bootp_packet_handler hook to try to do something with it. */

void dispatch ()
{
	struct timeval tv, *tvp;
	isc_result_t status;

	/* Wait for a packet or a timeout... XXX */
	do {
		tvp = process_outstanding_timeouts (&tv);
		status = omapi_one_dispatch (0, tvp);
	} while (status == ISC_R_TIMEDOUT || status == ISC_R_SUCCESS);
	log_fatal ("omapi_one_dispatch failed: %s -- exiting.",
		   isc_result_totext (status));
}

void add_timeout (when, where, what, ref, unref)
	TIME when;
	void (*where) PROTO ((void *));
	void *what;
	tvref_t ref;
	tvunref_t unref;
{
#ifdef LIBDHCP
	struct timeout t = {
		.when = when,
		.func = where,
		.what = what,
		.ref = ref,
		.unref = unref
	};
	struct timeout *new_timeouts;
	int i, pos = 0;
#else
	struct timeout *t, *q;
#endif

	/* See if this timeout supersedes an existing timeout. */
#ifdef LIBDHCP
	for (i = 0; i < ntimeouts; i++) {
		struct timeout *q = &timeouts[i];
#else
	t = (struct timeout *)0;
	for (q = timeouts; q; q = q -> next) {
		if ((where == NULL || q -> func == where) && q -> what == what) {
			if (t)
				t -> next = q -> next;
			else
				timeouts = q -> next;
			break;
		}
		t = q;
	}
#endif

#ifdef LIBDHCP
		/* If this one is already in the list with a different time,
		 * remove it and re-add */
		if ((where == NULL || q->func == where) &&
				q->what == what) {
			size_t n = (--ntimeouts - i) * sizeof (*q);
			memmove(&t, q, sizeof (t));

			if (n)
				memmove(&timeouts[i], &timeouts[i+1], n);

			if (ntimeouts) {
				new_timeouts = realloc(timeouts, ntimeouts * sizeof (*q));
				/* XXX broken API, no way to return error here */
				if (new_timeouts)
					timeouts = new_timeouts;
			} else {
				timeouts = NULL;
			}
			add_timeout(when, where, what, ref, unref);
#else
	/* If we didn't supersede a timeout, allocate a timeout
	   structure now. */
	if (!q) {
		if (free_timeouts) {
			q = free_timeouts;
			free_timeouts = q -> next;
		} else {
			q = ((struct timeout *)
				dmalloc (sizeof (struct timeout), MDL));
			if (!q)
				log_fatal ("add_timeout: no memory!");
		}
		memset (q, 0, sizeof *q);
		q -> func = where;
		q -> ref = ref;
		q -> unref = unref;
		if (q -> ref)
			(*q -> ref)(&q -> what, what, MDL);
		else
			q -> what = what;
	}

	q -> when = when;

	/* Now sort this timeout into the timeout list. */

	/* Beginning of list? */
	if (!timeouts || timeouts -> when > q -> when) {
		q -> next = timeouts;
		timeouts = q;
		return;
	}

	/* Middle of list? */
	for (t = timeouts; t -> next; t = t -> next) {
		if (t -> next -> when > q -> when) {
			q -> next = t -> next;
			t -> next = q;
#endif
			return;
#ifdef LIBDHCP
		} else if (timeouts[i].when > when) {
			pos = i;
#endif
		}
	}

#ifdef LIBDHCP
	/* If we didn't supersede an existing timeout, then pos is set
	 * to the timeout which will post after this one.  Insert this
	 * one before it. */

	new_timeouts = realloc(timeouts, sizeof (t) * (ntimeouts+1));
	/* XXX broken API, no way to return error here */
	if (new_timeouts) {
		/* ntimeouts = 10
		 * pos = 3;
		 * n = 10-3 * sizeof (t) = 7 * sizeof (t) 
		 */
		size_t n = (ntimeouts - pos) * sizeof (t);

		timeouts = new_timeouts;
		memmove(&timeouts[pos+1], &timeouts[pos], n);
		memmove(&timeouts[pos], &t, sizeof (t));
		ntimeouts++;
	}
#else
	/* End of list. */
	t -> next = q;
	q -> next = (struct timeout *)0;
#endif
}

void cancel_timeout (where, what)
	void (*where) PROTO ((void *));
	void *what;
{
#ifdef LIBDHCP
	struct timeout t;
	int i = 0;
#else
	struct timeout *t, *q;
#endif

	/* Look for this timeout on the list, and unlink it if we find it. */
#ifdef LIBDHCP
	for (i = 0; i < ntimeouts; i++) {
		struct timeout *new_timeouts, *q = &timeouts[i];

		if (q->func == where && q->what == what) {
			size_t n;

			memmove(&t, q, sizeof (t));

			n = (ntimeouts - i - 1) * sizeof (t);
			memmove(&timeouts[i+1], &timeouts[i], n);

			n = --ntimeouts * sizeof (t);
			new_timeouts = realloc(timeouts, n);
			/* XXX broken API, no way to return error here */
			if (new_timeouts || !n)
				timeouts = new_timeouts;

			if (t.unref)
				t.unref(t.what, MDL);
		}
	}
#else
	t = (struct timeout *)0;
	for (q = timeouts; q; q = q -> next) {
		if (q -> func == where && q -> what == what) {
			if (t)
				t -> next = q -> next;
			else
				timeouts = q -> next;
			break;
		}
		t = q;
	}

	/* If we found the timeout, put it on the free list. */
	if (q) {
		if (q -> unref)
			(*q -> unref) (&q -> what, MDL);
		q -> next = free_timeouts;
		free_timeouts = q;
	}
#endif
}

void cancel_all_timeouts ()
{
#ifdef LIBDHCP
	cur_time = TIME_MAX;
	process_outstanding_timeouts(NULL);
#else
	struct timeout *t, *n;
	for (t = timeouts; t; t = n) {
		n = t -> next;
		if (t -> unref && t -> what)
			(*t -> unref) (&t -> what, MDL);
		t -> next = free_timeouts;
		free_timeouts = t;
	}
#endif
}

void relinquish_timeouts ()
{
#ifdef LIBDHCP
	while (ntimeouts)
		cancel_timeout(timeouts[0].func, timeouts[0].what);
#else
	struct timeout *t, *n;
	for (t = free_timeouts; t; t = n) {
		n = t -> next;
		dfree (t, MDL);
	}
#endif
}
